@@ -20,14 +20,14 @@ import (
 	digest "github.com/opencontainers/go-digest"
 
 	"k8s.io/cli-runtime/pkg/genericclioptions"
-	"k8s.io/klog"
-	kcmdutil "k8s.io/kubernetes/pkg/kubectl/cmd/util"
-	"k8s.io/kubernetes/pkg/kubectl/util/templates"
+	"k8s.io/klog/v2"
+	kcmdutil "k8s.io/kubectl/pkg/cmd/util"
+	"k8s.io/kubectl/pkg/util/templates"
 
 	"github.com/openshift/library-go/pkg/image/dockerv1client"
-	imagereference "github.com/openshift/library-go/pkg/image/reference"
 	"github.com/openshift/library-go/pkg/image/registryclient"
 	"github.com/openshift/oc/pkg/cli/image/archive"
+	"github.com/openshift/oc/pkg/cli/image/imagesource"
 	imagemanifest "github.com/openshift/oc/pkg/cli/image/manifest"
 	"github.com/openshift/oc/pkg/cli/image/workqueue"
 )
@@ -56,30 +56,31 @@ var (
 		  [~<prefix>] - select the layer with the matching digest prefix or return an error
 
 		Negative indices are counted from the end of the list, e.g. [-1] selects the last
-		layer.`)
+		layer.
+		`)
 
 	example = templates.Examples(`
-# Extract the busybox image into the current directory
-%[1]s docker.io/library/busybox:latest
+		# Extract the busybox image into the current directory
+		oc image extract docker.io/library/busybox:latest
 
-# Extract the busybox image to a temp directory (must exist)
-%[1]s docker.io/library/busybox:latest --path /:/tmp/busybox
+		# Extract the busybox image to a temp directory (must exist)
+		oc image extract docker.io/library/busybox:latest --path /:/tmp/busybox
 
-# Extract a single file from the image into the current directory
-%[1]s docker.io/library/centos:7 --path /bin/bash:.
+		# Extract a single file from the image into the current directory
+		oc image extract docker.io/library/centos:7 --path /bin/bash:.
 
-# Extract all .repo files from the image's /etc/yum.repos.d/ folder.
-%[1]s docker.io/library/centos:7 --path /etc/yum.repos.d/*.repo:.
+		# Extract all .repo files from the image's /etc/yum.repos.d/ folder.
+		oc image extract docker.io/library/centos:7 --path /etc/yum.repos.d/*.repo:.
 
-# Extract the last layer in the image
-%[1]s docker.io/library/centos:7[-1]
+		# Extract the last layer in the image
+		oc image extract docker.io/library/centos:7[-1]
 
-# Extract the first three layers of the image
-%[1]s docker.io/library/centos:7[:3]
+		# Extract the first three layers of the image
+		oc image extract docker.io/library/centos:7[:3]
 
-# Extract the last three layers of the image
-%[1]s docker.io/library/centos:7[-3:]
-`)
+		# Extract the last three layers of the image
+		oc image extract docker.io/library/centos:7[-3:]
+	`)
 )
 
 type LayerInfo struct {
@@ -92,7 +93,7 @@ type LayerInfo struct {
 // an error, or false to stop processing.
 type TarEntryFunc func(*tar.Header, LayerInfo, io.Reader) (cont bool, err error)
 
-type Options struct {
+type ExtractOptions struct {
 	Mappings []Mapping
 
 	Files []string
@@ -108,6 +109,8 @@ type Options struct {
 	Confirm bool
 	DryRun  bool
 
+	FileDir string
+
 	genericclioptions.IOStreams
 
 	// ImageMetadataCallback is invoked once per image retrieved, and may be called in parallel if
@@ -122,8 +125,8 @@ type Options struct {
 	AllLayers bool
 }
 
-func NewOptions(streams genericclioptions.IOStreams) *Options {
-	return &Options{
+func NewExtractOptions(streams genericclioptions.IOStreams) *ExtractOptions {
+	return &ExtractOptions{
 		Paths: []string{},
 
 		IOStreams:       streams,
@@ -132,14 +135,14 @@ func NewOptions(streams genericclioptions.IOStreams) *Options {
 }
 
 // New creates a new command
-func New(name string, streams genericclioptions.IOStreams) *cobra.Command {
-	o := NewOptions(streams)
+func NewExtract(streams genericclioptions.IOStreams) *cobra.Command {
+	o := NewExtractOptions(streams)
 
 	cmd := &cobra.Command{
 		Use:     "extract",
 		Short:   "Copy files from an image to the filesystem",
 		Long:    desc,
-		Example: fmt.Sprintf(example, name+" extract"),
+		Example: example,
 		Run: func(c *cobra.Command, args []string) {
 			kcmdutil.CheckErr(o.Complete(c, args))
 			kcmdutil.CheckErr(o.Validate())
@@ -159,6 +162,7 @@ func New(name string, streams genericclioptions.IOStreams) *cobra.Command {
 	flag.BoolVarP(&o.PreservePermissions, "preserve-ownership", "p", o.PreservePermissions, "Preserve the permissions of extracted files.")
 	flag.BoolVar(&o.OnlyFiles, "only-files", o.OnlyFiles, "Only extract regular files and directories from the image.")
 	flag.BoolVar(&o.AllLayers, "all-layers", o.AllLayers, "For dry-run mode, process from lowest to highest layer and don't omit duplicate files.")
+	flag.StringVar(&o.FileDir, "dir", o.FileDir, "The directory on disk that file:// images will be copied under.")
 
 	return cmd
 }
@@ -173,7 +177,7 @@ type Mapping struct {
 	// Image is the raw input image to extract
 	Image string
 	// ImageRef is the parsed version of the raw input image
-	ImageRef imagereference.DockerImageReference
+	ImageRef imagesource.TypedImageReference
 	// LayerFilter can select which images to load
 	LayerFilter LayerFilter
 	// From is the directory or file in the image to extract
@@ -214,30 +218,39 @@ func parseMappings(images, paths, files []string, requireEmpty bool) ([]Mapping,
 			if len(mapping.From) > 0 {
 				mapping.From = strings.TrimPrefix(mapping.From, "/")
 			}
-			if len(mapping.To) > 0 {
-				fi, err := os.Stat(mapping.To)
-				if os.IsNotExist(err) {
-					return nil, fmt.Errorf("destination path does not exist: %s", mapping.To)
-				}
+
+			toPath := mapping.To
+			if len(toPath) == 0 {
+				toPath = "."
+			}
+			toPath, err := filepath.Abs(toPath)
+			if err != nil {
+				return nil, fmt.Errorf("cannot make path %q absolute: %v", mapping.To, err)
+			}
+			mapping.To = toPath
+
+			fi, err := os.Stat(mapping.To)
+			if os.IsNotExist(err) {
+				return nil, fmt.Errorf("destination path does not exist: %s", mapping.To)
+			}
+			if err != nil {
+				return nil, fmt.Errorf("invalid argument: %s", err)
+			}
+			if !fi.IsDir() {
+				return nil, fmt.Errorf("invalid argument: %s is not a directory", arg)
+			}
+			if requireEmpty {
+				f, err := os.Open(mapping.To)
 				if err != nil {
-					return nil, fmt.Errorf("invalid argument: %s", err)
+					return nil, fmt.Errorf("unable to check directory: %v", err)
 				}
-				if !fi.IsDir() {
-					return nil, fmt.Errorf("invalid argument: %s is not a directory", arg)
+				names, err := f.Readdirnames(1)
+				f.Close()
+				if err != nil && err != io.EOF {
+					return nil, fmt.Errorf("could not check for empty directory: %v", err)
 				}
-				if requireEmpty {
-					f, err := os.Open(mapping.To)
-					if err != nil {
-						return nil, fmt.Errorf("unable to check directory: %v", err)
-					}
-					names, err := f.Readdirnames(1)
-					f.Close()
-					if err != nil && err != io.EOF {
-						return nil, fmt.Errorf("could not check for empty directory: %v", err)
-					}
-					if len(names) > 0 {
-						return nil, fmt.Errorf("directory %s must be empty, pass --confirm to overwrite contents of directory", mapping.To)
-					}
+				if len(names) > 0 {
+					return nil, fmt.Errorf("directory %s must be empty, pass --confirm to overwrite contents of directory", mapping.To)
 				}
 			}
 			mappings = append(mappings, mapping)
@@ -260,11 +273,11 @@ func parseMappings(images, paths, files []string, requireEmpty bool) ([]Mapping,
 			}
 		}
 
-		src, err := imagereference.Parse(mapping.Image)
+		src, err := imagesource.ParseReference(mapping.Image)
 		if err != nil {
 			return nil, err
 		}
-		if len(src.Tag) == 0 && len(src.ID) == 0 {
+		if len(src.Ref.Tag) == 0 && len(src.Ref.ID) == 0 {
 			return nil, fmt.Errorf("source image must point to an image ID or image tag")
 		}
 		mapping.ImageRef = src
@@ -273,7 +286,7 @@ func parseMappings(images, paths, files []string, requireEmpty bool) ([]Mapping,
 	return mappings, nil
 }
 
-func (o *Options) Complete(cmd *cobra.Command, args []string) error {
+func (o *ExtractOptions) Complete(cmd *cobra.Command, args []string) error {
 	if err := o.FilterOptions.Complete(cmd.Flags()); err != nil {
 		return err
 	}
@@ -294,19 +307,24 @@ func (o *Options) Complete(cmd *cobra.Command, args []string) error {
 	return nil
 }
 
-func (o *Options) Validate() error {
+func (o *ExtractOptions) Validate() error {
 	if len(o.Mappings) == 0 {
 		return fmt.Errorf("you must specify one or more paths or files")
 	}
 	return o.FilterOptions.Validate()
 }
 
-func (o *Options) Run() error {
+func (o *ExtractOptions) Run() error {
 	ctx := context.Background()
 	fromContext, err := o.SecurityOptions.Context()
 	if err != nil {
 		return err
 	}
+	fromOptions := &imagesource.Options{
+		FileDir:         o.FileDir,
+		Insecure:        o.SecurityOptions.Insecure,
+		RegistryContext: fromContext,
+	}
 
 	stopCh := make(chan struct{})
 	defer close(stopCh)
@@ -316,29 +334,19 @@ func (o *Options) Run() error {
 			mapping := o.Mappings[i]
 			from := mapping.ImageRef
 			q.Try(func() error {
-				repo, err := fromContext.Repository(ctx, from.DockerClientDefaults().RegistryURL(), from.RepositoryName(), o.SecurityOptions.Insecure)
+				repo, err := fromOptions.Repository(ctx, from)
 				if err != nil {
-					return fmt.Errorf("unable to connect to image repository %s: %v", from.Exact(), err)
+					return fmt.Errorf("unable to connect to image repository %s: %v", from.String(), err)
 				}
 
-				srcManifest, location, err := imagemanifest.FirstManifest(ctx, from, repo, o.FilterOptions.Include)
+				srcManifest, location, err := imagemanifest.FirstManifest(ctx, from.Ref, repo, o.FilterOptions.Include)
 				if err != nil {
 					if imagemanifest.IsImageForbidden(err) {
-						var msg string
-						if len(o.Mappings) == 1 {
-							msg = "image does not exist or you don't have permission to access the repository"
-						} else {
-							msg = fmt.Sprintf("image %q does not exist or you don't have permission to access the repository", from)
-						}
+						msg := fmt.Sprintf("image %q does not exist or you don't have permission to access the repository", from)
 						return imagemanifest.NewImageForbidden(msg, err)
 					}
 					if imagemanifest.IsImageNotFound(err) {
-						var msg string
-						if len(o.Mappings) == 1 {
-							msg = "image does not exist"
-						} else {
-							msg = fmt.Sprintf("image %q does not exist", from)
-						}
+						msg := fmt.Sprintf("image %q does not exist", from)
 						return imagemanifest.NewImageNotFound(msg, err)
 					}
 					return fmt.Errorf("unable to read image %s: %v", from, err)
@@ -454,14 +462,14 @@ func (o *Options) Run() error {
 						if byEntry != nil {
 							cont, err := layerByEntry(r, options, info, byEntry, o.AllLayers, alreadySeen)
 							if err != nil {
-								err = fmt.Errorf("unable to iterate over layer %s from %s: %v", layer.Digest, from.Exact(), err)
+								err = fmt.Errorf("unable to iterate over layer %s from %s: %v", layer.Digest, from, err)
 							}
 							return cont, err
 						}
 
 						klog.V(4).Infof("Extracting layer %s with options %#v", layer.Digest, options)
 						if _, err := archive.ApplyLayer(mapping.To, r, options); err != nil {
-							return false, fmt.Errorf("unable to extract layer %s from %s: %v", layer.Digest, from.Exact(), err)
+							return false, fmt.Errorf("unable to extract layer %s from %s: %v", layer.Digest, from, err)
 						}
 						return true, nil
 					}()
